/********************************************************************************
 * Copyright (c) 2023 CEA-List
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 *
 ********************************************************************************/

#include "aidge/operator/Slice.hpp"

#include <cassert>
#include <cstddef>
#include <cstdint>
#include <string>
#include <utility>
#include <vector>

#include <fmt/format.h>

#include "aidge/backend/OperatorImpl.hpp"
#include "aidge/data/Data.hpp"
#include "aidge/data/Tensor.hpp"
#include "aidge/utils/ErrorHandling.hpp"
#include "aidge/utils/Types.h"

void Aidge::Slice_OpImpl::forward() {
    const Slice_Op& op = dynamic_cast<const Slice_Op&>(mOp);

    if (!op.getInput(0)) {
        AIDGE_THROW_OR_ABORT(std::runtime_error, "{}: input #0 should be associated with a Tensor", op.Type);
    }
    AIDGE_ASSERT((op.template getAttr<SliceAttr::Starts>().size() == op.template getAttr<SliceAttr::Ends>().size()) &&
                 (op.template getAttr<SliceAttr::Starts>().size() == op.template getAttr<SliceAttr::Axes>().size()),
                 "start, end and axes arguments should be the same size.");

    const std::size_t nbDims = op.getInput(0)->nbDims();

    const std::vector<std::size_t>& inputDims = op.getInput(0)->dims();
    auto outputDims = op.getInput(0)->dims();

    // compute index of the output's first element
    // compute output dimension at the same time (may change between two forward calls)
    std::size_t beginning = 0;
    const std::size_t nbAxes = op.template getAttr<SliceAttr::Axes>().size();
    for (std::size_t i = 0; i < nbAxes; ++i) {
        // For each slice operation get the params and cast them to size_t
        DimIdx_t axis = op.template getAttr<SliceAttr::Axes>()[i] >= 0 ?
                            static_cast<DimIdx_t>(op.template getAttr<SliceAttr::Axes>()[i]) :
                            static_cast<DimIdx_t>(op.template getAttr<SliceAttr::Axes>()[i] + static_cast<DimIdx_t>(inputDims.size()));
        DimSize_t start = op.template getAttr<SliceAttr::Starts>()[i] >= 0 ?
                            static_cast<DimSize_t>(op.template getAttr<SliceAttr::Starts>()[i]) :
                            static_cast<DimSize_t>(op.template getAttr<SliceAttr::Starts>()[i] + static_cast<DimSize_t>(inputDims[axis]));
        DimSize_t end = op.template getAttr<SliceAttr::Ends>()[i] >= 0 ?
                        static_cast<DimSize_t>(op.template getAttr<SliceAttr::Ends>()[i]) :
                        static_cast<DimSize_t>(op.template getAttr<SliceAttr::Ends>()[i] + static_cast<DimSize_t>(inputDims[axis]));
        const std::size_t stridePostAxis = std::accumulate(inputDims.cbegin()+axis+1, inputDims.cend(), std::size_t(1), std::multiplies<std::size_t>());
        beginning += start * stridePostAxis;
        const std::size_t sliceLength = end - start;
        outputDims[axis] = sliceLength;
    }
    op.getOutput(0)->resize(outputDims);


    // for inputDims = {4,5,5,3} & outputDims = {3,2,2,1}: substractDims = {1,5,5,3}
    std::vector<std::size_t> substractedDims = std::vector<std::size_t>(nbDims);
    for (std::size_t i = 0; i < nbDims; ++i) {
        substractedDims[i] = inputDims[i] - outputDims[i];
    }

    // for outputDims = {3,2,2,1}: prodOutputDims = {12,4,2,1}
    std::vector<std::size_t> prodOutputDims = std::vector<std::size_t>(nbDims);
    std::vector<std::size_t> prodInputDims = std::vector<std::size_t>(nbDims + 1);
    prodOutputDims[nbDims - 1] = outputDims[nbDims - 1];
    prodInputDims[nbDims - 1] = inputDims[nbDims - 1];
    prodInputDims[nbDims] = 1;
    for (std::size_t i = 2; i <= nbDims; ++i) {
        prodOutputDims[nbDims - i] = prodOutputDims[nbDims - i + 1] * outputDims[nbDims - i];
        prodInputDims[nbDims - i] = prodInputDims[nbDims - i + 1] * inputDims[nbDims - i];
    }

    std::size_t i = beginning;
    std::size_t size = 0; // number of elements to copy
    std::size_t offset = 0;
    for (std::size_t j = 0; j < prodOutputDims[0];) {
        ++size;
        ++i;
        ++j;
        bool newChunk = false;
        for (std::size_t idx = nbDims - 1; idx > 0; --idx) {
            if (j % prodOutputDims[idx] == 0) {
                i += substractedDims[idx] * prodInputDims[idx + 1];
                newChunk = true;
            }
        }

        if (newChunk) {
            op.getOutput(0)->getImpl()->copy(op.getInput(0)->getImpl()->rawPtr(beginning), size, offset);
            beginning = i;
            offset += size;
            size = 0;
        }
    }

    if (size > 0) {
        op.getOutput(0)->getImpl()->copy(op.getInput(0)->getImpl()->rawPtr(beginning), size, offset);
    }
}

const std::string Aidge::Slice_Op::Type = "Slice";

bool Aidge::Slice_Op::forwardDims(bool /*allowDataDependency*/) {
    // check inputs have been associated
    if (!getInput(0)) {
        AIDGE_THROW_OR_ABORT(std::runtime_error, "{}: input #0 should be associated with a Tensor", type());
    }

    if(!getInput(0)->empty())
    {
        if(this->template getAttr<SliceAttr::Starts>().empty() || this->template getAttr<SliceAttr::Ends>().empty() || this->template getAttr<SliceAttr::Axes>().empty())
        {
            if(getInput(1)->empty() || getInput(2)->empty() || getInput(3)->empty()) {
                AIDGE_THROW_OR_ABORT(std::runtime_error, "{}: Starts, Ends and Axes must be provided either as input or attributes", type());
            }

            AIDGE_ASSERT((mInputs[1]->dataType() == mInputs[2]->dataType()) && (mInputs[1]->dataType() == mInputs[3]->dataType()), "Slice inputs must have the same dataType.");

            this->template getAttr<SliceAttr::Starts>().clear(); // If both are provided input would override attrs
            this->template getAttr<SliceAttr::Starts>().reserve(getInput(1)->size());
            this->template getAttr<SliceAttr::Ends>().clear();
            this->template getAttr<SliceAttr::Ends>().reserve(getInput(1)->size());
            this->template getAttr<SliceAttr::Axes>().clear();
            this->template getAttr<SliceAttr::Axes>().reserve(getInput(1)->size());
            switch (mInputs[1]->dataType()) {
                case DataType::Float64:
                    std::copy_n(static_cast<double*>(mInputs[1]->getImpl()->rawPtr()),
                                getInput(1)->size(),
                                std::back_inserter(this->template getAttr<SliceAttr::Starts>()));
                    std::copy_n(static_cast<double*>(mInputs[2]->getImpl()->rawPtr()),
                                getInput(2)->size(),
                                std::back_inserter(this->template getAttr<SliceAttr::Ends>()));
                    std::copy_n(static_cast<double*>(mInputs[3]->getImpl()->rawPtr()),
                                getInput(3)->size(),
                                std::back_inserter(this->template getAttr<SliceAttr::Axes>()));
                    break;
                case DataType::Float32:
                    std::copy_n(static_cast<float*>(mInputs[1]->getImpl()->rawPtr()),
                                getInput(1)->size(),
                                std::back_inserter(this->template getAttr<SliceAttr::Starts>()));
                    std::copy_n(static_cast<float*>(mInputs[2]->getImpl()->rawPtr()),
                                getInput(2)->size(),
                                std::back_inserter(this->template getAttr<SliceAttr::Ends>()));
                    std::copy_n(static_cast<float*>(mInputs[3]->getImpl()->rawPtr()),
                                getInput(3)->size(),
                                std::back_inserter(this->template getAttr<SliceAttr::Axes>()));
                    break;
                case DataType::Int64:
                    std::copy_n(static_cast<std::int64_t*>(mInputs[1]->getImpl()->rawPtr()),
                                getInput(1)->size(),
                                std::back_inserter(this->template getAttr<SliceAttr::Starts>()));
                    std::copy_n(static_cast<std::int64_t*>(mInputs[2]->getImpl()->rawPtr()),
                                getInput(2)->size(),
                                std::back_inserter(this->template getAttr<SliceAttr::Ends>()));
                    std::copy_n(static_cast<std::int64_t*>(mInputs[3]->getImpl()->rawPtr()),
                                getInput(3)->size(),
                                std::back_inserter(this->template getAttr<SliceAttr::Axes>()));
                    break;
                case DataType::Int32:
                    std::copy_n(static_cast<std::int32_t*>(mInputs[1]->getImpl()->rawPtr()),
                                getInput(1)->size(),
                                std::back_inserter(this->template getAttr<SliceAttr::Starts>()));
                    std::copy_n(static_cast<std::int32_t*>(mInputs[2]->getImpl()->rawPtr()),
                                getInput(2)->size(),
                                std::back_inserter(this->template getAttr<SliceAttr::Ends>()));
                    std::copy_n(static_cast<std::int32_t*>(mInputs[3]->getImpl()->rawPtr()),
                                getInput(3)->size(),
                                std::back_inserter(this->template getAttr<SliceAttr::Axes>()));                                
                    break;
                default:
                    AIDGE_THROW_OR_ABORT(std::runtime_error, "{}: Indices input DataType is not supported.", type());
                    break;
            }
        }

        DimSize_t nbAxes = this->template getAttr<SliceAttr::Axes>().size();
        std::vector<DimSize_t> outDims = getInput(0)->dims();
        for (std::size_t i = 0; i < nbAxes; ++i) {
            DimIdx_t axis = this->template getAttr<SliceAttr::Axes>()[i] >= 0 ?
                            static_cast<DimIdx_t>(this->template getAttr<SliceAttr::Axes>()[i]) :
                            static_cast<DimIdx_t>(this->template getAttr<SliceAttr::Axes>()[i] + static_cast<DimIdx_t>(getInput(0)->nbDims()));
            DimSize_t start = this->template getAttr<SliceAttr::Starts>()[i] >= 0 ?
                              static_cast<DimSize_t>(this->template getAttr<SliceAttr::Starts>()[i]) :
                              static_cast<DimSize_t>(this->template getAttr<SliceAttr::Starts>()[i] + static_cast<DimSize_t>(getInput(0)->dims()[axis]));
            DimSize_t end = this->template getAttr<SliceAttr::Ends>()[i] >= 0 ?
                            static_cast<DimSize_t>(this->template getAttr<SliceAttr::Ends>()[i]) :
                            static_cast<DimSize_t>(this->template getAttr<SliceAttr::Ends>()[i] + static_cast<DimSize_t>(getInput(0)->dims()[axis]));

            const std::size_t sliceLength = end - start;
            // Check if slice length is valid
            if (sliceLength > getInput(0)->dims()[axis])
            {
                AIDGE_THROW_OR_ABORT(std::runtime_error, "ROI of Slice operator out of bounds");
            }
            outDims[axis] = sliceLength;
        }
        mOutputs[0]->resize(outDims);
        return true;
    }
    return false;
}

void Aidge::Slice_Op::setBackend(const std::string& name, Aidge::DeviceIdx_t device) {
    if (Registrar<Slice_Op>::exists({name})){
        SET_IMPL_MACRO(Slice_Op, *this, name);
    }
    else {
        mImpl = std::make_shared<Slice_OpImpl>(*this);
    }
    mOutputs[0]->setBackend(name, device);
}
